@@ -2,15 +2,17 @@ package AnyEvent::Open3::Simple;
 
 use strict;
 use warnings;
+use warnings::register;
 use AnyEvent;
 use IPC::Open3 qw( open3 );
 use Scalar::Util qw( reftype );
 use Symbol qw( gensym );
 use AnyEvent::Open3::Simple::Process;
 use Carp qw( croak );
+use File::Temp ();
 
-# ABSTRACT: interface to open3 under AnyEvent
-our $VERSION = '0.69_02'; # VERSION
+# ABSTRACT: Interface to open3 under AnyEvent
+our $VERSION = '0.79_02'; # VERSION
 
 
 sub new
@@ -20,75 +22,130 @@ sub new
   my $args = (reftype($_[0]) || '') eq 'HASH' ? shift : { @_ };
   my %self;
   $self{$_} = $args->{$_} || $default_handler for qw( on_stdout on_stderr on_start on_exit on_signal on_fail on_error on_success );
+  if($args->{stdin})
+  {
+    warnings::warnif('deprecated', "passing stdin to the constructor of AnyEvent::Open3::Simple is deprecated");
+    $self{stdin} = ref $args->{stdin} ? $args->{stdin} : \$args->{stdin};
+  }
   $self{impl} = $args->{implementation} 
              || $ENV{ANYEVENT_OPEN3_SIMPLE}
              || ($^O eq 'MSWin32' ? 'idle' : 'child');
   $self{raw} = $args->{raw} || 0;
-  unless($self{impl} =~ /^(idle|child)$/)
-  {
-    croak "unknown implementation $self{impl}";
-  }
+  croak "unknown implementation $self{impl}" unless $self{impl} =~ /^(idle|child)$/;
   bless \%self, $class;
 }
 
 
 sub run
 {
+  croak "run method requires at least one argument"
+    unless @_ >= 2;
+
+  my $proc_user = (ref $_[-1] eq 'CODE' ? pop : sub {});
+
+  my $stdin = $_[0]->{stdin};
+  $stdin = pop if ref $_[-1];
+
   my($self, $program, @arguments) = @_;
   
   my($child_stdin, $child_stdout, $child_stderr);
   $child_stderr = gensym;
+
+  local *TEMP;
+  if(defined $stdin)
+  {
+    my $file = File::Temp->new;
+    $file->autoflush(1);
+    $file->print(
+      ref($stdin) eq 'ARRAY'
+      ? join("\n", @{ $stdin })
+      : $$stdin
+    );
+    $file->seek(0,0);
+    open TEMP, '<&=', $file;
+    $child_stdin = '<&TEMP';
+  }
   
   my $pid = eval { open3 $child_stdin, $child_stdout, $child_stderr, $program, @arguments };
   
   if(my $error = $@)
   {
-    $self->{on_error}->($error);
+    $self->{on_error}->($error, $program, @arguments);
     return;
   }
   
   my $proc = AnyEvent::Open3::Simple::Process->new($pid, $child_stdin);
-  
-  $self->{on_start}->($proc);
-  
-  my $watcher_stdout = AnyEvent->io(
-    fh   => $child_stdout,
-    poll => 'r',
-    cb   => $self->{raw}
-    ? sub {
-        local $/;
-        $self->{on_stdout}->($proc, <$child_stdout>);
-      }
-    : sub {
+  $proc_user->($proc);
+
+  $self->{on_start}->($proc, $program, @arguments);
+
+  my $watcher_stdout;
+  my $watcher_stderr;
+
+  if($self->{raw})
+  {
+
+    require AnyEvent::Handle;
+    $watcher_stdout = AnyEvent::Handle->new(
+      fh => $child_stdout,
+      on_error => sub {
+        $self->{on_stdout}->($proc,$_[0]{rbuf});
+      },
+    );
+    $watcher_stdout->on_read(sub {
+      $watcher_stdout->push_read(sub {
+        $DB::single = 1;
+        $self->{on_stdout}->($proc,$_[0]{rbuf});
+        $_[0]{rbuf} = '';
+      });
+    });
+    $watcher_stderr = AnyEvent::Handle->new(
+      fh => $child_stderr,
+      on_error => sub {
+        $self->{on_stderr}->($proc,$_[0]{rbuf});
+      },
+    );
+    $watcher_stderr->on_read(sub {
+      $watcher_stdout->push_read(sub {
+        $self->{on_stderr}->($proc,$_[0]{rbuf});
+        $_[0]{rbuf} = '';
+      });
+    });
+
+  }
+  else
+  {
+
+    $watcher_stdout = AnyEvent->io(
+      fh   => $child_stdout,
+      poll => 'r',
+      cb   => sub {
         my $input = <$child_stdout>;
         return unless defined $input;
         $input =~ s/(\015?\012|\015)$//;
-        $self->{on_stdout}->($proc, $input);
-      }
-    ,
-  );
+        my $ref = $self->{on_stdout};
+        ref($ref) eq 'ARRAY' ? push @$ref, $input : $ref->($proc, $input);
+      },
+    );
 
-  my $watcher_stderr = AnyEvent->io(
-    fh   => $child_stderr,
-    poll => 'r',
-    cb   => $self->{raw}
-    ? sub {
-        local $/;
-        $self->{on_stderr}->($proc, <$child_stderr>);
-      }
-    : sub {
+    $watcher_stderr = AnyEvent->io(
+      fh   => $child_stderr,
+      poll => 'r',
+      cb   => sub {
         my $input = <$child_stderr>;
         return unless defined $input;
         $input =~ s/(\015?\012|\015)$//;
-        $self->{on_stderr}->($proc, $input);
-      }
-    ,
-  );
+        my $ref = $self->{on_stderr};
+        ref($ref) eq 'ARRAY' ? push @$ref, $input : $ref->($proc, $input);
+      },
+    );
+  }
 
   my $watcher_child;
 
   my $end_cb = sub {
-    my($pid, $status) = @_;
+    #my($pid, $status) = @_;
+    my $status = $_[1];
     my($exit_value, $signal) = ($status >> 8, $status & 127);
       
     $proc->close;
@@ -96,29 +153,30 @@ sub run
     # make sure we consume any stdout and stderr which hasn't
     # been consumed yet.  This seems to make on_out.t work on
     # cygwin
-    while(!eof $child_stdout)
+    if($self->{raw})
     {
-      if($self->{raw})
-      { local $/; $self->{on_stdout}->($proc,<$child_stdout>) }
-      else
+      local $/;
+      $self->{on_stdout}->($proc, scalar <$child_stdout>);
+      $self->{on_stderr}->($proc, scalar <$child_stderr>);
+    }
+    else
+    {
+      while(!eof $child_stdout)
       {
         my $input = <$child_stdout>;
         last unless defined $input;
         $input =~ s/(\015?\012|\015)$//;
-        $self->{on_stdout}->($proc,$input);
+        my $ref = $self->{on_stdout};
+        ref($ref) eq 'ARRAY' ? push @$ref, $input : $ref->($proc, $input);
       }
-    }
       
-    while(!eof $child_stderr)
-    {
-      if($self->{raw})
-      { local $/; $self->{on_stderr}->($proc,<$child_stderr>) }
-      else
+      while(!eof $child_stderr)
       {
         my $input = <$child_stderr>;
         last unless defined $input;
         $input =~ s/(\015?\012|\015)$//;
-        $self->{on_stderr}->($proc,$input);
+        my $ref = $self->{on_stderr};
+        ref($ref) eq 'ARRAY' ? push @$ref, $input : $ref->($proc, $input);
       }
     }
       
@@ -163,13 +221,15 @@ __END__
 
 =pod
 
+=encoding UTF-8
+
 =head1 NAME
 
-AnyEvent::Open3::Simple - interface to open3 under AnyEvent
+AnyEvent::Open3::Simple - Interface to open3 under AnyEvent
 
 =head1 VERSION
 
-version 0.69_02
+version 0.79_02
 
 =head1 SYNOPSIS
 
@@ -181,17 +241,19 @@ version 0.69_02
  
  my $ipc = AnyEvent::Open3::Simple->new(
    on_start => sub {
-     my $proc = shift; # isa AnyEvent::Open3::Simple::Process
+     my $proc = shift;       # isa AnyEvent::Open3::Simple::Process
+     my $program = shift;    # string
+     my @args = @_;          # list of arguments
      say 'child PID: ', $proc->pid;
    },
    on_stdout => sub { 
-     my $proc = shift; # isa AnyEvent::Open3::Simple::Process
-     my $line = shift; # string
+     my $proc = shift;       # isa AnyEvent::Open3::Simple::Process
+     my $line = shift;       # string
      say 'out: ', $string;
    },
    on_stderr => sub {
-     my $proc = shift; # isa AnyEvent::Open3::Simple::Process
-     my $line = shift; # string
+     my $proc = shift;       # isa AnyEvent::Open3::Simple::Process
+     my $line = shift;       # string
      say 'err: ', $line;
    },
    on_exit   => sub {
@@ -204,6 +266,8 @@ version 0.69_02
    },
    on_error => sub {
      my $error = shift;      # the exception thrown by IPC::Open3::open3
+     my $program = shift;    # string
+     my @args = @_;          # list of arguments
      warn "error: $error";
      $done->send;
    },
@@ -245,11 +309,21 @@ environment variable, like this:
 
  % export ANYEVENT_OPEN3_SIMPLE=idle
 
-=item * raw
+=item * stdin (DEPRECATED)
 
-If set to true (false is the default) then output will not be passed
-into the C<on_stdout> and C<on_stderr> callbacks as lines, but instead
-as chunks in whatever order they come.  New lines will not be stripped.
+Deprecated attribute for passing the entire content of stdin to the subprocess.
+This attribute may be removed from a future version of L<AnyEvent::Open3::Simple>
+(but not before 18 March 2015).  Instead use one of
+
+=over 4
+
+=item L<AnyEvent::Open3::Simple::Process#print>
+
+=item L<AnyEvent::Open3::Simple::Process#say>
+
+=item L<AnyEvent::Open3::Simple#run>
+
+=back
 
 =back
 
@@ -268,12 +342,16 @@ will be called.
 
 =over 4
 
-=item * C<on_start> ($proc)
+=item * C<on_start> ($proc, $program, @arguments)
 
 Called after the process is created, but before the run method returns
 (that is, it does not wait to re-enter the event loop first).
 
-=item * C<on_error> ($error)
+In versions 0.78 and better, this event also gets the program name
+and arguments passed into the L<run|AnyEvent::Open3::Simple#run>
+method.
+
+=item * C<on_error> ($error, $program, @arguments)
 
 Called when there is an execution error, for example, if you ask
 to run a program that does not exist.  No process is passed in
@@ -291,6 +369,10 @@ this will raise an C<on_error> event, on C<MSWin32> it will
 not trigger a C<on_error> and instead cause a normal exit
 with a exit value of 1.
 
+In versions 0.77 and better, this event also gets the program name
+and arguments passed into the L<run|AnyEvent::Open3::Simple#run>
+method.
+
 =item * C<on_stdout> ($proc, $line)
 
 Called on every line printed to stdout by the child process.
@@ -320,13 +402,46 @@ Called when the process returns a non-zero exit value.
 
 =head1 METHODS
 
-=head2 $ipc-E<gt>run($program, @arguments)
+=head2 run
+
+ $ipc->run($program, @arguments);
+ $ipc->run($program, @arguments, \$stdin);             # (version 0.76)
+ $ipc->run($program, @arguments, \@stdin);             # (version 0.76)
+ $ipc->run($program, @arguments, sub {...});           # (version 0.80)
+ $ipc->run($program, @arguments, \$stdin, sub {...});  # (version 0.80)
 
 Start the given program with the given arguments.  Returns
 immediately.  Any events that have been specified in the
 constructor (except for C<on_start>) will not be called until
 the process re-enters the event loop.
 
+You may optionally provide the full content of standard input
+as a string reference or list reference as the last argument
+(or second to last if you are providing a callback below).
+If provided as a list reference, it will be joined by new lines
+in whatever format is native to your Perl.  Currently on 
+(non cygwin) Windows (Strawberry, ActiveState) this is the only
+way to provide standard input to the subprocess.
+
+Do not mix the use of passing standard input to L<run|AnyEvent::Open3::Simple#run>
+and L<AnyEvent::Open3::Simple::Process#print> or L<AnyEvent::Open3::Simple::Process#say>,
+otherwise bad things may happen.
+
+In version 0.80 or better, you may provide a callback as the last argument
+which is called before C<on_start>, and takes the process object as its only 
+argument.  For example:
+
+ foreach my $i (1..10)
+ {
+   $ipc->run($prog, @args, \$stdin, sub {
+     my($proc) = @_;
+     $proc->user({ iteration => $i });
+   });
+ }
+
+This is useful for making data accessible to C<$ipc> object's callbacks that may
+be out of scope otherwise.
+
 =head1 CAVEATS
 
 Some AnyEvent implementations may not work properly with the method
@@ -344,14 +459,14 @@ attribute to force it use an idle watcher instead.  Patches for detecting
 environments where idle watchers should be used are welcome and
 encouraged.
 
-The pure perl implementation that comes with Perl
+The pure perl implementation that comes with L<AnyEvent>
 (L<AnyEvent::Impl::Perl>) does not seem to work with this module
 on Microsoft Windows so I make L<EV> a prereq on that platform 
-(which does work).
+(which is automatically used if installed and does work).
 
-Writing to a subprocesses stdin via L<AnyEvent::Open3::Simple::Process>'s
-C<print> method is unsupported on Microsoft Windows (it does work under
-Cygwin though).
+Writing to a subprocesses stdin with L<AnyEvent::Open3::Simple::Process#print>
+or L<AnyEvent::Open3::Simple::Process#say> is unsupported on Microsoft 
+Windows (it does work under Cygwin though).
 
 There are some traps for the unwary relating to buffers and deadlocks,
 L<IPC::Open3> is recommended reading.
@@ -400,6 +515,8 @@ contributors:
 
 Stephen R. Scaffidi
 
+Scott Wiersdorf
+
 =head1 COPYRIGHT AND LICENSE
 
 This software is copyright (c) 2012 by Graham Ollis.
